"""Try to find valid ISBNs inside a file or in `stdin` if no file was specified.

Searching for ISBNs in files uses progressively more resource-intensive methods
until some ISBNs are found, see the `documentation`_.

This is a Python port of `find-isbns.sh`_ from `ebook-tools`_ written in Shell
by `na--`_.

References
----------
* `ebook-tools`_

.. URLs

.. external links
.. _documentation: https://github.com/na--/ebook-tools#searching-for-isbns-in-files
.. _ebook-tools: https://github.com/na--/ebook-tools
.. _find-isbns.sh: https://github.com/na--/ebook-tools/blob/master/find-isbns.sh
.. _na--: https://github.com/na--
"""



#!/usr/bin/env python
"""This script is a Python port of `ebook-tools`_ written in Shell by `na--`_.

References
----------
* `ebook-tools`_

.. URLs

.. external links
.. _ebook-tools: https://github.com/na--/ebook-tools
.. _na--: https://github.com/na--
"""
import argparse

# import ipdb

import py_ebooktools
from py_ebooktools import edit, split_into_folders
from py_ebooktools.configs import default_config as default_cfg
from py_ebooktools.utils.genutils import (get_config_dict, init_log,
                                          namespace_to_dict,
                                          override_config_with_args, setup_log)

logger = init_log(__name__, __file__)

# ==============
# Default values
# ==============
_LOG_CFG = "log"
_MAIN_CFG = "main"


# Ref.: https://stackoverflow.com/a/14117511/14664104
def check_positive(value):
    try:
        # TODO: 2.0 rejected
        ivalue = int(value)
        if ivalue <= 0:
            raise argparse.ArgumentTypeError(
                f"{value} is an invalid positive int value")
    except ValueError:
        raise argparse.ArgumentTypeError(
            f"{value} is an invalid positive int value")
    else:
        return ivalue


def parse_edit_args(main_cfg):
    if main_cfg.reset:
        return edit.reset_file(main_cfg.cfg_type, main_cfg.app)
    else:
        return edit.edit_file(main_cfg.cfg_type, main_cfg.app)


def process_returned_values(returned_values):
    # ================================
    # Process previous returned values
    # ================================

    def log_opts_overriden(opts_overriden, msg, log_level='info'):
        nb_items = len(opts_overriden)
        for i, (cfg_name, old_v, new_v) in enumerate(opts_overriden):
            msg += "\t {}: {} --> {}".format(cfg_name, old_v, new_v)
            if i + 1 < nb_items:
                msg += "\n"
        getattr(logger, log_level)(msg)

    # Process 1st returned values: default args overriden by config options
    if returned_values.default_args_overriden:
        msg = "Default arguments overridden by config options:\n"
        log_opts_overriden(returned_values.default_args_overriden, msg)
    # Process 2nd returned values: config options overriden by args
    if returned_values.config_opts_overridden:
        msg = "Config options overridden by command-line arguments:\n"
        log_opts_overriden(returned_values.config_opts_overridden, msg, 'debug')
    # Process 3rd returned values: arguments not found in config file
    """
    if args_not_found_in_config:
        msg = "Command-line arguments not found in config file: " \
              "\n\t{}".format(args_not_found_in_config)
        logger.debug(msg)
    """


def setup_argparser():
    """Setup the argument parser for the command-line script.

    Returns
    -------
    parser : argparse.ArgumentParser
        Argument parser.

    """
    # Setup the parser
    parser = argparse.ArgumentParser(
        description='''\
This program is a Python port of ebook-tools written in Shell by na-- (See
https://github.com/na--/ebook-tools).

This program is a collection of tools for automated and
semi-automated organization and management of large ebook collections.

See subcommands below for a list of the tools that can be used.
''',
        formatter_class=argparse.RawDescriptionHelpFormatter)
    default_msg = " (default: {})"
    # ===============
    # General options
    # ===============
    # TODO: package name too? instead of program name
    parser.add_argument('--version', action='version',
                        version='%(prog)s v{}'.format(py_ebooktools.__version__))
    parser.add_argument("-q", "--quiet", action="store_true",
                        help="Enable quiet mode, i.e. nothing will be printed.")
    parser.add_argument("-v", "--verbose", action="store_true",
                        help='''Print various debugging information, e.g. print
                        traceback when there is an exception.''')
    parser.add_argument(
        "-d", "--dry-run", dest='dry_run', action="store_true",
        help='''If this is enabled, no file rename/move/symlink/etc.
        operations will actually be executed.''')
    parser.add_argument(
        "-r", "--reverse", dest='file_sort_reverse', action="store_true",
        help='''If this is enabled, the files will be sorted in reverse
        (i.e. descending) order. By default, they are sorted in ascending
        order.''')
    parser.add_argument(
        '--loglvl', dest='logging_level',
        choices=['debug', 'info', 'warning', 'error'],
        help='Set logging level for all loggers.'
             + default_msg.format(default_cfg.logging_level))
    parser.add_argument(
        '--logfmt', dest='logging_formatter',
        choices=['console', 'simple', 'only_msg'],
        help='Set logging formatter for all loggers.'
             + default_msg.format(default_cfg.logging_formatter))
    # =============================================
    # Options related to the input and output files
    # =============================================
    parser_input_output_group = parser.add_argument_group(
        title='Options related to the input and output files')
    parser_input_output_group.add_argument(
        '--oft', '--output-filename-template', dest='output_filename_template',
        metavar='TEMPLATE',
        help='''This specifies how the filenames of the organized files will
        look. It is a bash string that is evaluated so it can be very flexible
        (and also potentially unsafe).''' +
             default_msg.format(default_cfg.output_filename_template))
    parser_input_output_group.add_argument(
        '--ome', '--output-metadata-extension', dest='output_metadata_extension',
        metavar='EXTENSION',
        help='''If keep_metadata is enabled, this is the extension of the
        additional metadata file that is saved next to each newly renamed file.'''
             + default_msg.format(default_cfg.output_metadata_extension))
    # ===========
    # Subcommands
    # ===========
    subparsers = parser.add_subparsers(
        title='subcommands', description=None, dest='subcommand', required=True,
        help=None)
    # TODO: add aliases, see https://bit.ly/3s2fq87
    # ==========
    # Edit files
    # ==========
    # create the parser for the "edit" command
    parser_edit = subparsers.add_parser(
        'edit', help='''Edit a configuration file.''')
    parser_edit.add_argument(
        'cfg_type', choices=[_MAIN_CFG, _LOG_CFG],
        help='''The config file to edit which can either be the main
            configuration file ('{}') or the logging configuration file
            ('{}').'''.format(_MAIN_CFG, _LOG_CFG))
    group_edit = parser_edit.add_mutually_exclusive_group()
    group_edit.add_argument(
        '-a', '--app', metavar='NAME', nargs='?',
        help='''Name of the application to use for editing the config file. If
        no name is given, then the default application for opening this type of
        file will be used.''')
    group_edit.add_argument(
        "-r", "--reset", action="store_true",
        help='''Reset a configuration file ('config' or 'log') with factory
        default values.''')
    parser_edit.set_defaults(func=parse_edit_args)
    # ==============
    # Convert to txt
    # ==============
    # create the parser for the "convert" command
    parser_convert = subparsers.add_parser(
        'convert',
        help='''Convert the supplied file to a text file. It can optionally
        also use *OCR* for `.pdf`, `.djvu` and image files.''')
    # ==================
    # split-into-folders
    # ==================
    # create the parser for the "split-into-folders" command
    parser_split_into_folders = subparsers.add_parser(
        'split',
        help='''Split the supplied ebook files (and the accompanying metadata
        files if present) into folders with consecutive names that each contain
        the specified number of files.''')
    parser_split_into_folders.add_argument(
        'folder_with_books',
        help='''Folder with books which will be recursively scanned for files.
        The found files (and the accompanying metadata files if present) will
        be split into folders with consecutive names that each contain the
        specified number of files. The default value is the current working
        directory.''')
    parser_split_into_folders.add_argument(
        '-o', '--output-folder', dest='output_folder', metavar='PATH',
        help='''The output folder in which all the new consecutively named
        folders will be created. The default is the current working
        directory.''' + default_msg.format(default_cfg.output_folder))
    parser_split_into_folders.add_argument(
        '-s', '--start-number', dest='start_number', type=int,
        help='''The number of the first folder.'''
             + default_msg.format(default_cfg.start_number))
    parser_split_into_folders.add_argument(
        '-f', '--folder-pattern', dest='folder_pattern', metavar='PATTERN',
        help='''The print format string that specifies the pattern with which
        new folders will be created. By default it creates folders like
        00000000, 00001000, 00002000, .....'''
             + default_msg.format(default_cfg.folder_pattern).replace('%', '%%'))
    parser_split_into_folders.add_argument(
        '--fpf', '--files-per-folder', dest='files_per_folder', type=check_positive,
        help='''How many files should be moved to each folder.'''
             + default_msg.format(default_cfg.files_per_folder))
    parser_split_into_folders.set_defaults(func=split_into_folders.split)
    return parser


def main():
    try:
        parser = setup_argparser()
        args = parser.parse_args()
        # Get main cfg dict
        main_cfg = argparse.Namespace(**get_config_dict('main'))
        returned_values = override_config_with_args(main_cfg, parser)
        setup_log(main_cfg.quiet, main_cfg.verbose,
                  logging_level=main_cfg.logging_level,
                  logging_formatter=main_cfg.logging_formatter)
        # Override main configuration from file with command-line arguments
        process_returned_values(returned_values)
    except AssertionError as e:
        # TODO (IMPORTANT): use same logic as in Darth-Vader-RPi
        # TODO: add KeyboardInterruptError
        logger.error(e)
        return 1
    else:
        if args.subcommand == 'edit':
            return args.func(main_cfg)
        else:
            return args.func(**namespace_to_dict(main_cfg))


if __name__ == '__main__':
    # ebooktools --loglvl debug --logfmt simple split -o output_folder/ folder_with_books/
    # ebooktools --loglvl debug --logfmt simple split --fpf 3
    retcode = main()
    msg = "Program exited with {}".format(retcode)
    if retcode == 1:
        logger.error(msg)
    else:
        logger.info(msg)



"""Edit a configuration file, either the main configuration file (`main`) or the
logging configuration file (`log`).

The configuration file can be opened by a user-specified application (``app``) or
a default program associated with this type of file (when ``app`` is `None`).
"""
import platform
import os
import shutil

from py_ebooktools.utils.genutils import (get_config_filepath,
                                          get_logging_filepath,
                                          get_main_config_filepath,
                                          init_log, run_cmd)

logger = init_log(__name__, __file__)


# NOTE: https://stackoverflow.com/a/27163648 [launch PyCharm from terminal]
def edit_file(cfg_type, app=None):
    """Edit a configuration file.

    The user chooses what type of config file (`cfg_type`) to edit: 'log' for
    the `logging config file`_ and 'main' for the `main config file`_.

    The configuration file can be opened by a user-specified application (`app`)
    or a default program associated with this type of file (when `app` is
    :obj:`None`).

    Parameters
    ----------
    cfg_type : str, {'log', 'main'}
        The type of configuration file to edit. 'log' refers to the
        `logging config file`_ and 'main' to the `main config file`_.
    app : str, optional
        Name of the application to use for opening the config file, e.g.
        `TextEdit` (the default value is :obj:`None` which implies that the
        default application will be used to open the config file).

    Returns
    -------
    retcode : int
        If there is a `subprocess
        <https://docs.python.org/3/library/subprocess.html#subprocess.CalledProcessError>`_
        -related error, the return code is non-zero. Otherwise, it is 0 if the
        file can be successfully opened with an external program.

    """
    # Get path to the config file
    filepath = get_config_filepath(cfg_type)
    # Command to open the config file with the default application in the
    # OS or the user-specified app, e.g. `open filepath` in macOS opens the
    # file with the default app (e.g. atom)
    default_cmd_dict = {'Darwin': 'open {filepath}',
                        'Linux': 'xdg-open {filepath}',
                        'Windows': 'cmd /c start "" "{filepath}"'}
    # NOTE: check https://bit.ly/31htaOT (pymotw) for output from
    # platform.system on three OSes
    default_cmd = default_cmd_dict.get(platform.system())
    # NOTES:
    # - `app is None` implies that the default app will be used
    # - Otherwise, the user-specified app will be used
    cmd = default_cmd if app is None else app + " " + filepath
    retcode = 1
    result = None
    try:
        # IMPORTANT: if the user provided the name of an app, it will be used as
        # a command along with the file path, e.g. ``$ atom {filepath}``.
        # However, this case might not work if the user provided an app name
        # that doesn't refer to an executable, e.g. ``$ TextEdit {filepath}``
        # won't work. The failed case is further processed in the except block.
        result = run_cmd(cmd.format(filepath=filepath))
        retcode = result.returncode
    except FileNotFoundError:
        # This error happens if the name of the app can't be called as an
        # executable in the terminal
        # e.g. `TextEdit` can't be run in the terminal but `atom` can since the
        # latter refers to an executable.
        # To open `TextEdit` from the terminal, the command ``open -a TextEdit``
        # must be used on macOS.
        # TODO: IMPORTANT add the open commands for the other OSes
        specific_cmd_dict = {'Darwin': 'open -a {app}'.format(app=app)}
        # Get the command to open the file with the user-specified app
        cmd = specific_cmd_dict.get(platform.system(), app) + " " + filepath
        # TODO: explain DEVNULL, suppress stderr since we will display the error
        # TODO: IMPORTANT you might get a FileNotFoundError again?
        result = run_cmd(cmd)  # stderr=subprocess.DEVNULL)
        retcode = result.returncode
    if retcode == 0:
        logger.info("Opening the file {}...".format(
            os.path.basename(filepath)))
        logger.debug(f"Filepath: {filepath}")
    else:
        if result:
            err = result.stderr.decode().strip()
            logger.error(err)
    return retcode


def reset_file(cfg_type, app=None, create_if_not=False):
    # Get path to the config file
    filepath = get_config_filepath(cfg_type)
    logger.info("Resetting the file {}...".format(
        os.path.basename(filepath)))
    logger.debug(f"Filepath: {filepath}")
    # Copy it from the default one
    if cfg_type == 'main':
        src = get_main_config_filepath(default_config=True)
    else:
        src = get_logging_filepath(default_config=True)
    shutil.copy(src, filepath)
    logger.info("File was reset!")
    return 0



"""General utilities
"""
import codecs
import json
import logging.config
import os
import shlex
import shutil
import subprocess
import sys
from argparse import Namespace
from collections import namedtuple, OrderedDict
from logging import NullHandler
from runpy import run_path
from types import SimpleNamespace

import py_ebooktools
from py_ebooktools.configs import default_config as default_cfg
from py_ebooktools.utils.logutils import (init_log, set_logging_field_width,
                                          set_logging_formatter, set_logging_level)

logger = init_log(__name__, __file__)
logger.addHandler(NullHandler())

CFG_TYPES = ['main', 'log']
CONFIGS_DIRNAME = 'configs'


def get_config_dict(cfg_type='main'):
    return load_cfg_dict(get_config_filepath(cfg_type), cfg_type)


def get_config_filepath(cfg_type='main'):
    if cfg_type == 'main':
        cfg_filepath = get_main_config_filepath()
    elif cfg_type == 'log':
        cfg_filepath = get_logging_filepath()
    else:
        raise ValueError(f"Invalid cfg_type: {cfg_type}")
    return cfg_filepath


def get_settings(conf, cfg_type):
    if cfg_type == 'log':
        set_logging_field_width(conf['logging'])
        return conf['logging']
    elif cfg_type == 'main':
        _settings = {}
        for opt_name, opt_value in conf.items():
            if opt_name.startswith('__') and opt_name.endswith('__'):
                continue
            elif isinstance(opt_value, type(os)):
                # e.g. import config
                continue
            else:
                _settings.setdefault(opt_name, opt_value)
        return _settings
    else:
        raise ValueError(f"Invalid cfg_type: {cfg_type}")


def load_cfg_dict(cfg_filepath, cfg_type):

    def _load_cfg_dict(cfg_filepath, cfg_type):
        if file_ext == '.py':
            cfg_dict = run_path(cfg_filepath)
            cfg_dict = get_settings(cfg_dict, cfg_type)
        elif file_ext == '.json':
            cfg_dict = load_json(cfg_filepath)
        else:
            raise TypeError("Config file extension not supported: "
                            f"{cfg_filepath}")
        return cfg_dict

    assert cfg_type in CFG_TYPES, f"Invalid cfg_type: {cfg_type}"
    _, file_ext = os.path.splitext(cfg_filepath)
    try:
        cfg_dict = _load_cfg_dict(cfg_filepath, cfg_type)
    except FileNotFoundError as e:
        print(f"WARNING: Config file '{os.path.basename(cfg_filepath)}' will "
              "be created")
        # Copy it from the default one
        # TODO: IMPORTANT destination with default?
        if cfg_type == 'main':
            src = get_main_config_filepath(default_config=True)
        else:
            src = get_logging_filepath(default_config=True)
        shutil.copy(src, cfg_filepath)
        cfg_dict = _load_cfg_dict(cfg_filepath, cfg_type)
    return cfg_dict


def load_json(filepath, encoding='utf8'):
    """Load JSON data from a file on disk.

    If using Python version betwee 3.0 and 3.6 (inclusive), the data is
    returned as :obj:`collections.OrderedDict`. Otherwise, the data is
    returned as :obj:`dict`.

    Parameters
    ----------
    filepath : str
        Path to the JSON file which will be read.
    encoding : str, optional
        Encoding to be used for opening the JSON file in read mode (the default
        value is '*utf8*').

    Returns
    -------
    data : dict or collections.OrderedDict
        Data loaded from the JSON file.

    Raises
    ------
    OSError
        Raised if any I/O related error occurs while reading the file, e.g. the
        file doesn't exist.

    References
    ----------
    `Are dictionaries ordered in Python 3.6+? (stackoverflow)`_

    """
    try:
        with codecs.open(filepath, 'r', encoding) as f:
            if sys.version_info.major == 3 and sys.version_info.minor <= 6:
                data = json.load(f, object_pairs_hook=OrderedDict)
            else:
                data = json.load(f)
    except OSError:
        raise
    else:
        return data


def mkdir(path):
    # Since path can be relative to the cwd
    path = os.path.abspath(path)
    dirname = os.path.basename(path)
    if os.path.exists(path):
        logger.debug(f"Folder already exits: {path}")
        logger.debug(f"Skipping it!")
    else:
        logger.debug(f"Creating folder '{dirname}': {path}")
        os.mkdir(path)
        logger.debug("Folder created!")


def move(src, dest):
    # Since path can be relative to the cwd
    src = os.path.abspath(src)
    filename = os.path.basename(src)
    if os.path.exists(dest):
        logger.debug(f"File already exits: '{filename}'")
        logger.debug(f"Destination folder path: {os.path.dirname(dest)}")
        logger.debug(f"Skipping it!")
    else:
        logger.debug(f"Moving '{filename}'")
        logger.debug(f"Destination folder path: {os.path.dirname(dest)}")
        shutil.move(src, dest)
        logger.debug("File moved!")


def namespace_to_dict(ns):
    namspace_classes = [Namespace, SimpleNamespace]
    # TODO: check why not working anymore
    # if isinstance(ns, SimpleNamespace):
    if type(ns) in namspace_classes:
        adict = vars(ns)
    else:
        adict = ns
    for k, v in adict.items():
        # if isinstance(v, SimpleNamespace):
        if type(v) in namspace_classes:
            v = vars(v)
            adict[k] = v
        if isinstance(v, dict):
            namespace_to_dict(v)
    return adict


def override_config_with_args(config, parser):
    ignored_args = ['func', 'subparser_name']
    # If config is Namespace
    config = vars(config)
    args = parser.parse_args().__dict__
    # args_not_found_in_config = []
    results = namedtuple("results", "config_opts_overridden default_args_overriden")
    results.config_opts_overridden = []
    results.default_args_overriden = []
    for arg_name, arg_val in args.items():
        if arg_name in ignored_args:
            continue
        config_val = config.get(arg_name)
        # No value was specified, use default value
        default_val = getattr(default_cfg, arg_name, 'not_found')
        if arg_val is not None:
            # User specified a value in the command-line
            if arg_val != config_val:
                config[arg_name] = arg_val
                results.config_opts_overridden.append(
                    (arg_name, config_val, arg_val))
            # else: command-line arg and config option same value, nothing to do
        elif config_val is not None:
            # User provided a value from the config file
            if config_val != default_val:
                results.default_args_overriden.append(
                    (arg_name, default_val, config_val))
        else:
            if default_val != 'not_found':
                config[arg_name] = default_val
            else:
                raise AttributeError("No value could be found for the "
                                     f"argument {arg_name}")
    return results


def run_cmd(cmd):
    """Run a shell command with arguments.

    The shell command is given as a string but the function will split it in
    order to get a list having the name of the command and its arguments as
    items.

    Parameters
    ----------
    cmd : str
        Command to be executed, e.g. ::

            open -a TextEdit text.txt

    Returns
    -------
    retcode: int
        Returns code which is 0 if the command was successfully completed.
        Otherwise, the return code is non-zero.

    Raises
    ------
    FileNotFoundError
        Raised if the command ``cmd`` is not recognized, e.g.
        ``$ TextEdit {filepath}`` since `TextEdit` is not an executable.

    """
    try:
        if sys.version_info.major == 3 and sys.version_info.minor <= 6:
            # TODO: PIPE not working as arguments and capture_output new in
            # Python 3.7
            # Ref.: https://stackoverflow.com/a/53209196
            #       https://bit.ly/3lvdGlG
            result = subprocess.run(shlex.split(cmd))
        else:
            result = subprocess.run(shlex.split(cmd), capture_output=True)
    except FileNotFoundError:
        raise
    else:
        return result


def setup_log(quiet=False, verbose=False, logging_level=None,
              logging_formatter=None):
    package_path = os.getcwd()
    log_filepath = get_logging_filepath()
    main_cfg_msg = f"Main config path: {get_main_config_filepath()}"
    main_log_msg = f'Logging path: {log_filepath}'
    # Get logging cfg dict
    log_dict = load_cfg_dict(log_filepath, cfg_type='log')
    # NOTE: if quiet and verbose are both activated, only quiet will have an effect
    # TODO: get first cfg_dict to setup log (same in train_models.py)
    if not quiet:
        if verbose:
            # verbose supercedes logging_level
            set_logging_level(log_dict, level='DEBUG')
        else:
            if logging_level:
                logging_level = logging_level.upper()
                # TODO: add console_for_users at the top
                set_logging_level(log_dict, level=logging_level)
        if logging_formatter:
            set_logging_formatter(log_dict, formatter=logging_formatter)
        # Load logging config dict
        logging.config.dictConfig(log_dict)
    # =============
    # Start logging
    # =============
    logger.info("Running {} v{}".format(py_ebooktools.__name__,
                                        py_ebooktools.__version__))
    logger.info("Verbose option {}".format(
        "enabled" if verbose else "disabled"))
    logger.debug("Working directory: {}".format(package_path))
    logger.debug(main_cfg_msg)
    logger.debug(main_log_msg)


# -------------------------------
# Configs: dirpaths and filepaths
# -------------------------------
def get_configs_dirpath():
    from py_ebooktools.configs import __path__
    return __path__[0]


def get_logging_filepath(default_config=False):
    # TODO: add names of logging files as global
    if default_config:
        return os.path.join(get_configs_dirpath(), 'default_logging.py')
    else:
        return os.path.join(get_configs_dirpath(), 'logging.py')


def get_main_config_filepath(default_config=False):
    # TODO: add names of config files as global
    if default_config:
        return os.path.join(get_configs_dirpath(), 'default_config.py')
    else:
        return os.path.join(get_configs_dirpath(), 'config.py')



"""Convert the supplied file to a text file.

It can optionally also use *OCR* for `.pdf`, `.djvu` and image files.

This is a Python port of `convert-to-txt.sh`_ from `ebook-tools`_ written in
Shell by `na--`_.

References
----------
* `ebook-tools`_

.. URLs

.. external links
.. _convert-to-txt.sh: https://github.com/na--/ebook-tools/blob/master/convert-to-txt.sh
.. _ebook-tools: https://github.com/na--/ebook-tools
.. _na--: https://github.com/na--
"""



"""Options for the script ebooktools

ebooktools is a Python port of the shell scripts developed by na-- for
organizing and managing ebook collections. See
https://github.com/na--/ebook-tools.

The options are described in full detail in the project's README @
https://github.com/raul23/python-ebook-tools

IMPORTANT: these options can also be passed to the script ebooktools via
command-line parameters (Run `ebooktools -h` for a detailed list of the options).
Also, command-line parameters supersede the variables defined in this
configuration file. Most parameters are not required and if nothing is
specified, the default value will be used.

The options are separated based on sections:
1. General options
  1.1 General control flags
  1.2 Options related to extracting ISBNs from files and finding metadata by ISBN
  1.3 Options for OCR
  1.4 Options related to extracting and searching for non-ISBN metadata
  1.5 Options related to the input and output files
  1.6 Miscellaneous options
2. Script options
  2.1 edit-config-files
  2.2 organize-ebooks
    2.2.1 Specific options for organizing files
    2.2.2 Output options
  2.3 interactive-organizer
  2.4 find-isbns
  2.5 convert-to-txt
  2.6 rename-calibre-library
  2.7 split-into-folders
"""
import os

# ==================
# 1. General options
# ==================
# All of these options are part of the common library and may affect some or
# all of the scripts.

# 1.1 General control flags
# =========================
quiet = False
verbose = False
dry_run = False
# symlink_only = False
# keep_metadata = False
logging_level = 'info'
logging_formatter = 'only_msg'

# 1.2 Options related to extracting ISBNs from files and finding metadata by ISBN
# ===============================================================================
# isbn_regex = ?
# isbn_blacklist_regex = '^(0123456789|([0-9xX])\2{9})$'
# isbn_direct_grep_files = '^text/(plain|xml|html)$'
# isbn_ignored_files = ?
# isbn_grep_reorder_files = True
# isbn_grep_rf_scan_first = 400
# isbn_grep_rf_reverse = 50
# reorder-files-for-grep = (isbn_grep_reorder_files, isbn_grep_rf_scan_first, isbn_grep_rf_reverse
# NOTE: If you use Calibre versions that are older than 2.84, it's required to
# manually set the following option to an empty string.
# metadata_fetch_order = ['Goodreads', 'Amazon.com', 'Google', 'ISBNDB', 'WorldCat xISBN', 'OZON.ru']

# 1.3 Options for OCR
# ===================
# ocr_enabled = False
# ocr_only_first_last_pages = (7, 3)
# ocr_command = 'tesseract_wrapper'

# 1.4 Options related to extracting and searching for non-ISBN metadata
# =====================================================================
# token_min_length = 3
# tokens_to_ignore = ?
# organize_without_isbn_sources = ['Goodreads' ,'Amazon.com', 'Google']

# 1.5 Options related to the input and output files
# =================================================
output_filename_template = "${d[AUTHORS]// & /, } - ${d[SERIES]:+[${d[SERIES]}] " \
                           "- }${d[TITLE]/:/ -}${d[PUBLISHED]:+ (${d[PUBLISHED]%%-*})}" \
                           "${d[ISBN]:+ [${d[ISBN]}]}.${d[EXT]}"
# If keep_metadata is enabled, this is the extension of the additional
# metadata file that is saved next to each newly renamed file.
output_metadata_extension = 'meta'

# 1.6 Miscellaneous options
# =========================
# file_sort_flags = []
file_sort_reverse = False
# debug_prefix_length = 40

# =================
# 2. Script options
# =================

# 2.1 edit-config-files
# =====================
# Name of the application to use for editing the config file.
# If no name is given, then the default application for opening this type of
# file will be used.
app = None

# 2.2 organize-ebooks
# ===================
# 2.2.1 Specific options for organizing files
# -------------------------------------------
# corruption_check_only = False
# tested_archive_extensions = '^(7z|bz2|chm|arj|cab|gz|tgz|gzip|zip|rar|xz|tar|epub|docx|odt|ods|cbr|cbz|maff|iso)$'
# organize_without_isbn = False
# without_isbn_ignore = ?
# pamphlet_included_files = '\.(png|jpg|jpeg|gif|bmp|svg|csv|pptx?)$'
# pamphlet_excluded_files = '\.(chm|epub|cbr|cbz|mobi|lit|pdb)$'
# pamphlet_max_pdf_pages = 50
# pamphlet_max_filesize_kb = 250

# 2.2.2 Output options
# --------------------
# output_folder = os.getcwd()
# If organize_without_isbn is enabled, this is the folder to which all ebooks
# that were renamed based on non-ISBN metadata will be moved to.
# output_folder_uncertain = None
# output_folder_corrupt = None
# output_folder_pamphlets = None

# 2.3 interactive-organizer
# =========================
# output_folder = []
# quick_mode = False
# custom_move_base_dir = None
# restore_original_base_dir = None
# diacritic_difference_masking = ?
# match_partial_words = False

# 2.4 find-isbns
# ==============
# Some general options affect this script (especially the ones related to
# extracting ISBNs from files, see above section 1.4).
# isbn_return_separator = "$'\n'"

# 2.5 convert-to-txt
# ==================
# There are no local options, but some of the general options affect this
# script's behavior a lot, especially the OCR ones (see '1.3 Options for OCR').

# 2.6 rename-calibre-library
# ==========================
# output_folder = os.getcwd()
# save_metadata = 'recreate'

# 2.7 split-into-folders
# ======================
output_folder = os.getcwd()
start_number = 0
folder_pattern = '%05d000'
files_per_folder = 1000



logging = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters":
    {
        "console":
        {
          "format": "%(name)-{auto_field_width}s | %(levelname)-8s | %(message)s"
        },
        "console_time":
        {
          "format": "%(asctime)s | %(levelname)-8s | %(message)s"
        },
        "only_msg":
        {
          "format": "%(message)s"
        },
        "simple":
        {
          "format": "%(levelname)-8s %(message)s"
        },
        "simple2":
        {
          "format": "%(levelname)-8s | %(message)s"
        },
        "verbose":
        {
          "format": "%(asctime)s | %(name)-{auto_field_width}s | %(levelname)-8s | %(message)s"
        }
    },

    "handlers":
    {
        "console":
        {
          "level": "WARNING",
          "class": "logging.StreamHandler",
          "formatter": "only_msg"
        },
        "console_only_msg":
        {
          "level": "INFO",
          "class": "logging.StreamHandler",
          "formatter": "only_msg"
        },
        "file":
        {
          "level": "INFO",
          "class": "logging.FileHandler",
          "filename": "debug.log",
          "mode": "a",
          "formatter": "simple",
          "delay": True
        }
    },

    "loggers":
    {
        # --------------------------------------
        # Loggers using console_only_msg handler
        # --------------------------------------
        "data":
        {
          "level": "INFO",
          "handlers": ["console_only_msg"],
          "propagate": False
        },
        # -----------------------------
        # Loggers using console handler
        # -----------------------------
        "py_ebooktools.edit":
        {
          "level": "DEBUG",
          "handlers": ["console"],
          "propagate": False
        },
        "py_ebooktools.split_into_folders":
        {
          "level": "DEBUG",
          "handlers": ["console"],
          "propagate": False
        },
        "scripts.ebooktools":
        {
          "level": "DEBUG",
          "handlers": ["console"],
          "propagate": False
        },
        "utils.genutils":
        {
          "level": "DEBUG",
          "handlers": ["console"],
          "propagate": False
        },
    },

    "root":
    {
        "level": "INFO",
        "handlers": ["console"],
        "propagate": False
    }
}



import logging
import os
import sys
from logging import NullHandler


# TODO: explain cases
def get_logger_name(module__name__, module___file__, package_name=None):
    if os.path.isabs(module___file__):
        # e.g. initcwd or editcfg
        module_name = os.path.splitext(os.path.basename(module___file__))[0]
        package_path = os.path.dirname(module___file__)
        package_name = os.path.basename(package_path)
        logger_name = "{}.{}".format(
            package_name,
            module_name)
    elif module__name__ == '__main__' or not module__name__.count('.'):
        # e.g. train_models.py or explore_data.py
        if package_name is None:
            package_name = os.path.basename(os.getcwd())
        logger_name = "{}.{}".format(
            package_name,
            os.path.splitext(module___file__)[0])
    elif module__name__.count('.') > 1:
        logger_name = '.'.join(module__name__.split('.')[-2:])
    else:
        # e.g. importing mlutils from train_models.py
        logger_name = module__name__
    return logger_name


def init_log(module__name__, module___file__=None, package_name=None):
    if module___file__:
        logger_ = logging.getLogger(get_logger_name(module__name__,
                                                    module___file__,
                                                    package_name))
    elif module__name__.count('.') > 1:
        logger_name = '.'.join(module__name__.split('.')[-2:])
        logger_ = logging.getLogger(logger_name)
    else:
        logger_ = logging.getLogger(module__name__)
    logger_.addHandler(NullHandler())
    return logger_


# TODO: specify log_dict change inline
def set_logging_field_width(log_dict):
    names = log_dict['loggers'].keys()
    if sys.argv and os.path.basename(sys.argv[0]) == 'ebooktools':
        names = [n for n in names if not n.startswith('default_')]
    size_longest_name = len(max(names, key=len))
    for k, v in log_dict['formatters'].items():
        try:
            # TODO: add auto_field_width at the top
            v['format'] = v['format'].format(auto_field_width=size_longest_name)
        except KeyError:
            continue


def set_logging_formatter(log_dict, handler_names=None, formatter='simple'):
    # TODO: assert handler_names and formatter
    """
    for handler_name in handler_names:
        log_dict['handlers'][handler_name]['formatter'] = formatter
    """
    handler_names = handler_names if handler_names else []
    for handler_name, handler_val in log_dict['handlers'].items():
        if not handler_names or handler_name in handler_names:
            handler_val['formatter'] = formatter


def set_logging_level(log_dict, handler_names=None, logger_names=None,
                      level='DEBUG'):
    # TODO: assert handler_names, logger_names and level
    handler_names = handler_names if handler_names else []
    logger_names = logger_names if logger_names else []
    keys = ['handlers', 'loggers']
    for k in keys:
        for name, val in log_dict[k].items():
            if (not handler_names and not logger_names) or \
                    (k == 'handlers' and name in handler_names) or \
                    (k == 'loggers' and name in logger_names):
                val['level'] = level



"""setup.py file for the package ``py_ebooktools``.

The PyPi project name is ``py-ebooktools`` and the package name is
``py_ebooktools``.

"""
import fnmatch
import os
import sys
from setuptools import find_packages, setup
from setuptools.command.build_py import build_py as build_py_orig

from py_ebooktools import __version__, __test_version__


excluded = ['py_ebooktools/configs/config.py',
            'py_ebooktools/configs/logging.py']


# IMPORTANT: bdist_wheel behaves differently to sdist
# - MANIFEST.in works for source distributions, but it's ignored for wheels,
#   See https://bit.ly/3s2Kt3p
# - "bdist_wheel would always include stuff that I'd manage to exclude from the sdist"
#   See https://bit.ly/3t7SsO4
# Code reference: https://stackoverflow.com/a/56044136/14664104
class build_py(build_py_orig):
    def find_package_modules(self, package, package_dir):
        modules = super().find_package_modules(package, package_dir)
        return [
            (pkg, mod, file)
            for (pkg, mod, file) in modules
            if not any(fnmatch.fnmatchcase(file, pat=pattern) for pattern in excluded)
        ]


# Choose the correct version based on script's arg
if len(sys.argv) > 1 and sys.argv[1] == "testing":
    VERSION = __test_version__
    # Remove "testing" from args so setup doesn't process "testing" as a cmd
    sys.argv.remove("testing")
else:
    VERSION = __version__

# Directory of this file
dirpath = os.path.abspath(os.path.dirname(__file__))

# The text of the README file (used on PyPI)
# NOTE: encoding for py3.6 and less, see https://stackoverflow.com/a/49131427
with open(os.path.join(dirpath, "README_pypi.rst"), encoding="utf-8") as f:
    README = f.read()

# The text of the requirements.txt file
# TODO: empty for now
with open(os.path.join(dirpath, "requirements.txt")) as f:
    REQUIREMENTS = f.read().splitlines()


setup(name='py-ebooktools',
      version=VERSION,
      description='''Program for organizing and managing ebook collections. It 
      is a Python port of ebook-tools (shell scripts).''',
      long_description=README,
      long_description_content_type='text/x-rst',
      classifiers=[
        'Development Status :: 3 - Alpha',  # TODO: change version to beta or stable
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities'
      ],
      keywords='ebook-manager calibre script',
      url='https://github.com/raul23/py-ebooktools',
      author='Raul C.',
      author_email='rchfe23@gmail.com',
      license='GPLv3',
      packages=find_packages(exclude=['tests']),
      cmdclass={'build_py': build_py},
      include_package_data=True,
      install_requires=REQUIREMENTS,
      entry_points={
        'console_scripts': ['ebooktools=py_ebooktools.scripts.ebooktools:main']
      },
      project_urls={  # Optional
          'Bug Reports': 'https://github.com/raul23/py-ebooktools/issues',
          # TODO: 'Documentation': 'https://?.readthedocs.io/',
          'Source': 'https://github.com/raul23/py-ebooktools',
      },
      zip_safe=False)



"""Library that has useful functions for building other ebook management tools.

This is a Python port of `lib.sh`_ from `ebook-tools`_ written in Shell by
`na--`_.

.. URLs

.. external links
.. _ebook-tools: https://github.com/na--/ebook-tools
.. _lib.sh: https://github.com/na--/ebook-tools/blob/master/lib.sh
.. _na--: https://github.com/na--
"""



"""Split the supplied ebook files (and the accompanying metadata files if
present) into folders with consecutive names that each contain the specified
number of files.

This is a Python port of `split-into-folders.sh`_ from `ebook-tools`_ written
in Shell by `na--`_.

References
----------
* `ebook-tools`_

.. URLs

.. external links
.. _ebook-tools: https://github.com/na--/ebook-tools
.. _na--: https://github.com/na--
.. _split-into-folders.sh: https://github.com/na--/ebook-tools/blob/master/split-into-folders.sh
"""
import math
import os
# import ipdb
from pathlib import Path

from py_ebooktools.configs import default_config as default_cfg
from py_ebooktools.utils.genutils import init_log, mkdir, move

logger = init_log(__name__, __file__)


def split(folder_with_books=Path.cwd(),
          folder_pattern=default_cfg.folder_pattern,
          start_number=default_cfg.start_number,
          output_folder=default_cfg.output_folder,
          files_per_folder=default_cfg.files_per_folder,
          output_metadata_extension=default_cfg.output_metadata_extension,
          dry_run=default_cfg.dry_run,
          reverse=default_cfg.file_sort_reverse, **kwargs):
    files = []
    for fp in Path(folder_with_books).rglob('*'):
        # File extension
        # ipdb.set_trace()
        ext = fp.suffix.split('.')[-1]
        # Ignore directory, metadata and hidden files
        if Path.is_file(fp) and ext != output_metadata_extension and \
                not fp.name.startswith('.'):
            # TODO: debug logging
            # print(fp)
            files.append((fp))
        # TODO: debug logging skip directory/file
    logger.info("Files sorted {}".format("in desc" if reverse else "in asc"))
    files.sort(key=lambda x: x.name, reverse=reverse)
    current_folder_num = start_number
    start_index = 0
    # Get width of zeros for folder format pattern
    left, right = folder_pattern.split('%')[-1].split('d')
    width = int(left) + len(right)
    total_files = len(files)
    logger.info(f"Total number of files to be split into folders: {total_files}")
    logger.info(f"Number of files per folder: {files_per_folder}")
    number_splits = math.ceil(total_files / files_per_folder)
    logger.info(f"Number of splits: {number_splits}")
    logger.info("Starting splits...")
    while True:
        if start_index >= len(files):
            # TODO: debug logging
            logger.info(f"End of splits!")
            break
        chunk = files[start_index:start_index+files_per_folder]
        start_index += files_per_folder
        logger.debug(f"Found {len(chunk)} number of files...")
        current_folder_basename = '{0:0{width}}'.format(
            current_folder_num, width=width)
        current_folder = os.path.join(output_folder, current_folder_basename)
        current_folder_metadata = os.path.join(
            output_folder, current_folder_basename + '.' + output_metadata_extension)
        current_folder_num += 1
        logger.debug(f"Creating folders '{current_folder}' and "
                     f"'{current_folder_metadata}'...")
        if not dry_run:
            mkdir(current_folder)
            mkdir(current_folder_metadata)
        logger.debug(f"Moving files...")
        if not dry_run:
            for file_to_move in chunk:
                file_dest = os.path.join(current_folder, file_to_move.name)
                move(file_to_move, file_dest)
                # Move metadata file if found
                metadata_name = f'{file_to_move.stem}.{output_metadata_extension}'
                metada_file_to_move = file_to_move.parent.joinpath(metadata_name)
                if metada_file_to_move.exists():
                    logger.debug(f"Found metadata file: {metada_file_to_move}")
                    metadata_dest = os.path.join(current_folder_metadata,
                                                 metadata_name)
                    move(metada_file_to_move, metadata_dest)
    return 0
